UNIT lexical_analyser;

  INTERFACE

    USES standard, wincrt;

    TYPE  symbol      = (sym_process,  sym_var,        sym_in,      sym_out,
                         sym_send,     sym_receive,    sym_to,      sym_from,
                         sym_any,      sym_deref,      sym_if,      sym_then,
                         sym_else,     sym_while,      sym_do,      sym_begin,
                         sym_end,
                         sym_becomes,
                         sym_semi,     sym_comma,      sym_dot,
                         sym_open,     sym_close,
                         sym_plus,     sym_minus,      sym_multiply, sym_div,  sym_mod,
                         sym_or,       sym_and,        sym_not,
                         sym_eq,       sym_neq,
                         sym_ls,       sym_gtr,        sym_leq,      sym_geq,
                         sym_constant, sym_Bool_const,
                         sym_id,       sym_eof,        sym_error
                        );

    PROCEDURE open_source(name: string; VAR ok: Boolean);

    PROCEDURE get_symbol;

    FUNCTION  symbol_kind: symbol;

    FUNCTION  id_value: string;

    FUNCTION  const_value: natural;

    FUNCTION  Bool_value: Boolean;

    PROCEDURE close_source;

    PROCEDURE report_error(message: string);

  IMPLEMENTATION

    CONST id_size      = 8;

          max_word     = 25;

    TYPE  small_letter = 'a'..'z';

          big_letter   = 'A'..'Z';

          word_count   = 0..max_word;

          word_index   = 1..max_word;

          id_string    = STRING [id_size];

          word_entry   = RECORD
                           word: id_string;
                           sym : symbol
                         END;

    VAR   source_ch    : char;

          id           : id_string;

          sym          : symbol;

          Bool         : Boolean;

          constant     : natural;

          lower        : SET OF small_letter;

          upper        : SET OF big_letter;

          id_chars     : SET OF char;

          digits       : SET OF char;

          singles      : SET OF char;

          source       : text;

          words        : word_count;

          word_table   : ARRAY [word_index] OF word_entry;

          source_object: string;
          source_opened: Boolean;

    PROCEDURE get_source;
    BEGIN
      IF eoln(source) THEN
      BEGIN
        writeln(source_ch);
        read(source, source_ch);
        read(source, source_ch);
        source_ch:=' '
      END
      ELSE
      BEGIN
        write(source_ch);
        read(source, source_ch)
      END
    END;

    PROCEDURE add_word(w: id_string; s: symbol);
    BEGIN
    END;

    PROCEDURE set_up_word_table;
    BEGIN
      words:=0;
      add_word('SEND',    sym_send   );
      add_word('RECEIVE', sym_receive);
      add_word('FROM',    sym_from   );
      add_word('TO',      sym_to     );
      add_word('ANY',     sym_any    );
      add_word('BEGIN',   sym_begin  );
      add_word('END',     sym_end    );
      add_word('IF',      sym_if     );
      add_word('THEN',    sym_then   );
      add_word('ELSE',    sym_else   );
      add_word('WHILE',   sym_while  );
      add_word('DO',      sym_do     );
      add_word('VAR',     sym_var    );
      add_word('PROCESS', sym_process);
      add_word('IN',      sym_in     );
      add_word('OUT',     sym_out    );
      add_word('DIV',     sym_div    );
      add_word('MOD',     sym_mod    );
      add_word('NOT',     sym_not    );
      add_word('AND',     sym_and    );
      add_word('OR',      sym_or     )
    END;

    PROCEDURE skip_layout;
    BEGIN
      WHILE (source_ch = ' ') AND NOT eof(source) DO get_source
    END;

    PROCEDURE get_word(VAR w: id_string);
    BEGIN
      w:='';
      WHILE (source_ch IN upper) DO
      BEGIN
        IF length(w) < id_size THEN w:=concat(w, source_ch);
        get_source
      END
    END;

    PROCEDURE look_up_word(w: id_string; VAR s: symbol);
    BEGIN
    END;

    PROCEDURE get_word_sym;
        VAR word: id_string;
    BEGIN
      get_word(word);
      look_up_word(word, sym)
    END;

    PROCEDURE get_id_sym;
    BEGIN
      id:='';
      WHILE source_ch IN id_chars DO
      BEGIN
        IF (length(id) < id_size) AND (source_ch <> '_') THEN id:=concat(id, source_ch);
        get_source
      END;
      IF (id = 'true') OR (id = 'false') THEN
      BEGIN
        Bool:=(id = 'true'); sym:=sym_Bool_const
      END
      ELSE sym:=sym_id
    END;

    PROCEDURE get_constant;
        VAR digit: natural;
    BEGIN
      sym:=sym_constant;
      constant:=0;
      WHILE source_ch IN digits DO
      BEGIN
        digit:=ord(source_ch) - ord('0');
        IF constant <= (maxint - digit) DIV 10 THEN constant:=10*constant + digit
        ELSE sym:=sym_error;
        get_source
      END
    END;

    PROCEDURE get_symbol;
    BEGIN
      skip_layout;
      IF eof(source)       THEN sym:=sym_eof
      ELSE
      IF source_ch IN upper  THEN get_word_sym
      ELSE
      IF source_ch IN lower  THEN get_id_sym
      ELSE
      IF source_ch IN digits THEN get_constant
      ELSE
      IF source_ch IN singles THEN
      BEGIN
        CASE source_ch OF
          ';': sym:=sym_semi;
          ',': sym:=sym_comma;
          '.': sym:=sym_dot;
          '+': sym:=sym_plus;
          '-': sym:=sym_minus;
          '*': sym:=sym_multiply;
          '(': sym:=sym_open;
          ')': sym:=sym_close;
          '=': sym:=sym_eq;
          '^': sym:=sym_deref
        END;
        get_source
      END
      ELSE
      IF source_ch IN [':', '>', '<'] THEN
      BEGIN
        CASE source_ch OF
          ':': BEGIN
                 get_source;
                 IF source_ch = '=' THEN
                 BEGIN
                   sym:=sym_becomes; get_source
                 END
                 ELSE sym:=sym_error
               END;
          '>': BEGIN
                 get_source;
                 IF source_ch = '=' THEN
                 BEGIN
                   sym:=sym_geq; get_source
                 END
                 ELSE sym:=sym_gtr
               END;
          '<': BEGIN
                 get_source;
                 IF source_ch = '=' THEN
                 BEGIN
                   sym:=sym_leq; get_source
                 END
                 ELSE
                 IF source_ch = '>' THEN
                 BEGIN
                   sym:=sym_neq; get_source
                 END
                 ELSE sym:=sym_ls
               END
        END
      END
      ELSE
      BEGIN
        sym:=sym_error; get_source
      END
    END;

    PROCEDURE init_lexical_analysis;
    BEGIN
      lower   :=['a'..'z'];
      upper   :=['A'..'Z'];
      digits  :=['0'..'9'];
      id_chars:=lower + digits + ['_'];
      singles :=[';', ',', '.', '+', '-', '*', '(', ')', '=', '^'];
      set_up_word_table;
      source_ch:=' ';
      get_symbol
    END;

    PROCEDURE open_source(* name: string; VAR ok: Boolean *);
       VAR filemode: integer;
    BEGIN
      {$I-}
        Assign(source, name);
        FileMode := 0;
        Reset(source);
      {$I+}
      ok:= (IOResult = 0) and (name <> '');
      source_object:=name; source_opened:=true;
      IF ok THEN init_lexical_analysis
    END;

    FUNCTION  symbol_kind(*: symbol *);
    BEGIN
      symbol_kind:=sym
    END;

    FUNCTION  id_value(*: string *);
    BEGIN
      id_value:=id
    END;

    FUNCTION  const_value(*: natural *);
    BEGIN
      const_value:=constant
    END;

    FUNCTION  Bool_value(*: Boolean *);
    BEGIN
      Bool_value:=Bool
    END;

    PROCEDURE close_source;
    BEGIN
      close(source); source_opened:=false
    END;

    PROCEDURE report_error(* message: string *);
    BEGIN
    END;

END.


