UNIT semantic_analyser;

  INTERFACE

    USES wincrt, standard, abstract_syntax, syntax_analyser,
         lexical_analyser;

    PROCEDURE check_program(VAR p: a_program; VAR ok: Boolean);

  IMPLEMENTATION

    TYPE type_kind = (undefined_type, integer_type, Boolean_type);

    VAR  processes   : process_list;
         inputs      : process_list;
         outputs     : process_list;
         variables   : var_list;
         process_vars: var_list;

    PROCEDURE a_semantic_error(message: string; VAR ok: Boolean);
        VAR ch: char;
    BEGIN
      report_error(message);
      writeln;
      writeln(message, '.');
      writeln('Push Accept to abort...'); ch:=readkey;
      ok:=false
    END;

    PROCEDURE check_statement(s: a_statement; VAR ok: Boolean); FORWARD;

    PROCEDURE chk_proc_ref(VAR r: a_proc_ref; processes: process_list; VAR ok: Boolean);
        VAR id: identifier;
    BEGIN
      CASE r.kind OF
        named_ref : BEGIN
                      id:=r.the_ref^.name;
                      r.the_ref:=process_ref(id, processes);
                      IF r.the_ref = nil THEN
                        a_semantic_error(concat('Process ', id, ' not in connection set'), ok);
                      ok:=true
                    END;
        proc_deref: ok:=true;
        any_ref   : ok:=true
      END
    END;

    PROCEDURE check_expression(e: an_expression; VAR ok: Boolean; VAR expr_type: type_kind);

        PROCEDURE check_binary(e: an_expression; VAR ok: Boolean; VAR expr_type: type_kind);

            VAR type_1, type_2: type_kind;

            FUNCTION op_string(operator: a_binary_op): string;
            BEGIN
              CASE operator OF
                add_op: op_string:='+'  ;   sub_op: op_string:='-'  ;  mult_op: op_string:='*';
                div_op: op_string:='DIV';   mod_op: op_string:='MOD';
                or_op : op_string:='OR' ;   and_op: op_string:='AND';
                eq_op : op_string:='='  ;   neq_op: op_string:='<>' ;  ls_op  : op_string:='<';
                leq_op: op_string:='<=' ;   gtr_op: op_string:='>'  ;  geq_op : op_string:='>='
              END
            END;

            PROCEDURE check_operator(operator: a_binary_op; type_1, type_2: type_kind;
                                     VAR ok: Boolean; VAR result_type: type_kind
                                    );
            BEGIN
              CASE operator OF
                add_op, sub_op, mult_op, div_op, mod_op:
                  BEGIN
                    result_type:=integer_type;
                    IF (type_1 <> integer_type) OR (type_2 <> integer_type) THEN
                      a_semantic_error(concat('The operands of ', op_string(operator), ' must be of type integer'), ok)
                  END;
                or_op, and_op:
                  BEGIN
                    result_type:=Boolean_type;
                    IF (type_1 <> Boolean_type) OR (type_2 <> Boolean_type) THEN
                      a_semantic_error(concat('The operands of ', op_string(operator), ' must be of type Boolean'), ok)
                  END;
                eq_op, neq_op, ls_op, leq_op, gtr_op, geq_op:
                  BEGIN
                    result_type:=Boolean_type;
                    IF (type_1 <> integer_type) OR (type_2 <> integer_type) THEN
                      a_semantic_error(concat('The operands of ', op_string(operator), ' must be of type integer'), ok)
                  END
              END
            END;

        BEGIN
          WITH e^.the_binary DO
          BEGIN
            check_expression(operand_1, ok, type_1);
            IF ok THEN
            BEGIN
              check_expression(operand_2, ok, type_2);
              IF ok THEN check_operator(operator, type_1, type_2, ok, expr_type)
            END
          END
        END;

        PROCEDURE check_unary(e: an_expression; VAR ok: Boolean; VAR expr_type: type_kind);

            VAR type_1: type_kind;

            FUNCTION op_string(operator: a_unary_op): string;
            BEGIN
              CASE operator OF
                plus_op : op_string:='+';
                minus_op: op_string:='-';
                not_op  : op_string:='NOT'
              END
            END;

            PROCEDURE check_operator(operator: a_unary_op; type_1: type_kind; VAR ok: Boolean; VAR result_type: type_kind);
            BEGIN
              CASE operator OF
                plus_op, minus_op:
                  BEGIN
                    result_type:=integer_type;
                    IF type_1 <> integer_type THEN
                      a_semantic_error(concat('The operand of ', op_string(operator), ' must be of type integer'), ok)
                  END;
                not_op:
                  BEGIN
                    result_type:=Boolean_type;
                    IF type_1 <> Boolean_type THEN
                      a_semantic_error(concat('The operand of ', op_string(operator), ' must be of type Boolean'), ok)
                  END
              END
            END;

        BEGIN
          WITH e^.the_unary DO
          BEGIN
            check_expression(operand, ok, type_1);
            IF ok THEN check_operator(operator, type_1, ok, expr_type)
          END
        END;

        PROCEDURE check_io(e: an_expression; VAR ok: Boolean; VAR expr_type: type_kind);
        BEGIN
          expr_type:=Boolean_type;
          check_statement(e^.the_io, ok)
        END;

        PROCEDURE check_var(e: an_expression; VAR ok: Boolean; VAR expr_type: type_kind);
        BEGIN
          expr_type:=integer_type;
          ok:=true
        END;

    BEGIN
      ok:=true;
      CASE e^.kind OF
        const_expr     : expr_type:=integer_type;
        Bool_const_expr: expr_type:=Boolean_type;
        var_expr       : check_var   (e, ok, expr_type);
        binary_expr    : check_binary(e, ok, expr_type);
        unary_expr     : check_unary (e, ok, expr_type);
        io_expr        : check_io    (e, ok, expr_type)
      END
    END;

    PROCEDURE check_stmt_list(l: stmt_list; VAR ok: Boolean); FORWARD;

    PROCEDURE check_assignment(s: a_statement; VAR ok: Boolean);
        VAR expr_type: type_kind;
    BEGIN
      WITH s^.the_assignment DO
      BEGIN
        check_expression(value, ok, expr_type);
        IF ok THEN
          IF expr_type <> integer_type THEN
            a_semantic_error('<expression> in assignment must be an integer expression', ok)
      END
    END;

    PROCEDURE check_send(s: a_statement; VAR ok: Boolean);
        VAR expr_type: type_kind;
    BEGIN
      WITH s^.the_send DO
      BEGIN
        IF value = nil THEN ok:=true
        ELSE
        IF value^.kind = gen_expr THEN ok:=true
        ELSE
        BEGIN
          check_expression(value, ok, expr_type);
          IF ok THEN
            IF expr_type <> integer_type THEN
              a_semantic_error('SEND <expression> must be an integer expression', ok)
        END;
        IF ok THEN chk_proc_ref(destination, outputs, ok)
      END
    END;

    PROCEDURE check_receive(s: a_statement; VAR ok: Boolean);
        VAR source_id: identifier;
    BEGIN
      WITH s^.the_receive DO
        chk_proc_ref(source, inputs, ok)
    END;

    PROCEDURE check_if(s: a_statement; VAR ok: Boolean);
        VAR expr_type: type_kind;
    BEGIN
      WITH s^.the_if DO
      BEGIN
        check_expression(condition, ok, expr_type);
        IF ok THEN
          IF expr_type <> Boolean_type THEN
            a_semantic_error('IF <expression> must be a Boolean expression', ok);
        IF ok THEN check_statement(then_part, ok);
        IF ok THEN check_statement(else_part, ok)
      END
    END;

    PROCEDURE check_while(s: a_statement; VAR ok: Boolean);
        VAR expr_type: type_kind;
    BEGIN
      WITH s^.the_while DO
      BEGIN
        check_expression(condition, ok, expr_type);
        IF ok THEN
          IF expr_type <> Boolean_type THEN
            a_semantic_error('WHILE <expression> must be a Boolean expression', ok);
        IF ok THEN check_statement(statement, ok)
      END
    END;

    PROCEDURE check_compound(s: a_statement; VAR ok: Boolean);
    BEGIN
      check_stmt_list(s^.the_compound, ok)
    END;

    PROCEDURE check_statement(* s: a_statement; VAR ok: Boolean *);
    BEGIN
      CASE s^.kind OF
        empty_stmt     : ok:=true;
        assignment_stmt: check_assignment(s, ok);
        while_stmt     : check_while     (s, ok);
        if_stmt        : check_if        (s, ok);
        send_stmt      : check_send      (s, ok);
        receive_stmt   : check_receive   (s, ok);
        compound_stmt  : check_compound  (s, ok)
      END
    END;

    PROCEDURE check_stmt_list(* l: stmt_list; VAR ok: Boolean *);
    BEGIN
      ok:=true;
      WHILE ok AND (l <> nil) DO
      BEGIN
        check_statement(l^.head, ok);
        l:=l^.tail
      END
    END;

    PROCEDURE check_var_dec(variables: var_list; VAR ok: Boolean);
        VAR v: a_variable;
    BEGIN
      ok:=true;
      WHILE ok AND (variables <> nil) DO
      BEGIN
        v:=variables^.head;
        IF var_ref(v^.name, variables^.tail) <> nil THEN
          a_semantic_error(concat('variable ', v^.name, ' declared twice'), ok);
        variables:=variables^.tail
      END
    END;

    PROCEDURE check_connections(l: process_list; VAR ok: Boolean);
        VAR id: identifier;
    BEGIN
      ok:=true;
      WHILE ok AND (l <> nil) DO
      BEGIN
        id:=l^.head^.name;
        l^.head:=process_ref(id, processes);
        IF l^.head = nil THEN
          a_semantic_error(concat('Undefined process <identifier> in connection list ', id), ok);
        l:=l^.tail
      END
    END;

    PROCEDURE chk_process(p: a_process; VAR ok: Boolean);
    BEGIN
      check_connections(p^.inputs, ok); inputs:=p^.inputs;
      IF ok THEN
      BEGIN
        check_connections(p^.outputs, ok); outputs:=p^.outputs;
        IF ok THEN
        BEGIN
          variables   :=p^.variables;
          process_vars:=p^.process_vars;
          check_var_dec(variables, ok);
          IF ok THEN check_stmt_list(p^.body, ok)
        END
      END
    END;

    PROCEDURE check_unique_processes(processes: process_list; VAR ok: Boolean);
        VAR p: a_process;
    BEGIN
      ok:=true;
      WHILE ok AND (processes <> nil) DO
      BEGIN
        p:=processes^.head;
        IF process_ref(p^.name, processes^.tail) <> nil THEN
          a_semantic_error(concat('PROCESS ', p^.name, ' declared twice'), ok);
        processes:=processes^.tail
      END
    END;

    PROCEDURE add_special_processes(VAR processes: process_list);

        VAR screen  : a_process;
            keyboard: a_process;

        FUNCTION cons_process(p: a_process; l: process_list): process_list;
            VAR new_list: process_list;
        BEGIN
          new(new_list);
          new_list^.head:=p;
          new_list^.tail:=l;
          cons_process  :=new_list
        END;

    BEGIN
      screen  :=new_process('screen',   nil, nil, nil, nil, nil);
      keyboard:=new_process('keyboard', nil, nil, nil, nil, nil);
      processes:=cons_process(keyboard, cons_process(screen, processes))
    END;

    PROCEDURE check_program(* VAR p: a_program; VAR ok: Boolean *);
        CONST for_UIMS = false;
        VAR   l        : process_list;
    BEGIN
      IF for_UIMS THEN
      BEGIN
        (*
        wi_new_window(w, 0, 2, 65, 7, true);
        wi_flash_string('Semantic Analysis...'); writeln;
        writeln
        *)
      END;
        l:=p;
        add_special_processes(p);
        check_unique_processes(p, ok);
        processes:=p;
        WHILE ok AND (l <> nil) DO
        BEGIN
          writeln('PROCESS ', l^.head^.name);
          chk_process(l^.head, ok);
          IF NOT ok THEN write(chr(7));
          l:=l^.tail
        END;
      IF for_UIMS THEN (* wi_dispose_window(true)*)
    END;

END.
