UNIT syntax_analyser;

  INTERFACE

    USES wincrt, abstract_syntax;

    PROCEDURE rec_program(name: string; VAR ok: Boolean; VAR prog: a_program);

  IMPLEMENTATION

    USES lexical_analyser;

    TYPE symbol_set     = SET OF symbol;

    VAR  first_statement  : symbol_set;
         first_factor     : symbol_set;
         first_term       : symbol_set;
         first_simple_expr: symbol_set;
         first_expression : symbol_set;

         variables        : var_list;
         process_vars     : var_list;

    PROCEDURE init_syntax_analysis;
    BEGIN
      first_statement  :=[sym_id, sym_while, sym_if, sym_send, sym_receive, sym_begin];
      first_factor     :=[sym_id, sym_constant, sym_Bool_const, sym_open, sym_not, sym_send, sym_receive];
      first_term       :=first_factor;
      first_simple_expr:=first_term + [sym_plus, sym_minus];
      first_expression :=first_simple_expr
    END;

    PROCEDURE syntax_error(message: string; VAR ok: Boolean);
        VAR ch: char;
    BEGIN
      report_error(message);
      writeln;
      writeln;
      writeln(message, '.');
      writeln('Push Return to abort...'); ch:=readkey;
      ok:=false
    END;

    FUNCTION make_variable(id: identifier; VAR ok: Boolean): a_variable;
        VAR v: a_variable;
    BEGIN
      v:=var_ref(id, variables);
      IF v = nil THEN syntax_error(concat('undeclared variable ', id), ok);
      make_variable:=v
    END;

    FUNCTION make_proc_var(id: identifier): a_variable;
        VAR v: a_variable;
    BEGIN
      v:=var_ref(id, process_vars);
      IF v = nil THEN
      BEGIN
        v:=new_variable(id);
        add_variable(v, process_vars)
      END;
      make_proc_var:=v
    END;

    FUNCTION binary_operator(symbol_kind: symbol): a_binary_op;
    BEGIN
      CASE symbol_kind OF
        sym_plus    : binary_operator:=add_op;
        sym_minus   : binary_operator:=sub_op;
        sym_multiply: binary_operator:=mult_op;
        sym_div     : binary_operator:=div_op;
        sym_mod     : binary_operator:=mod_op;
        sym_or      : binary_operator:=or_op;
        sym_and     : binary_operator:=and_op;
        sym_eq      : binary_operator:=eq_op;
        sym_neq     : binary_operator:=neq_op;
        sym_ls      : binary_operator:=ls_op;
        sym_leq     : binary_operator:=leq_op;
        sym_gtr     : binary_operator:=gtr_op;
        sym_geq     : binary_operator:=geq_op
      END
    END;

    FUNCTION unary_operator(symbol_kind: symbol): a_unary_op;
    BEGIN
      CASE symbol_kind OF
        sym_plus : unary_operator:=plus_op;
        sym_minus: unary_operator:=minus_op;
        sym_not  : unary_operator:=not_op
      END
    END;

    PROCEDURE rec_statement(VAR ok: Boolean; VAR s: a_statement); FORWARD;

    PROCEDURE rec_expression(VAR ok: Boolean; VAR e: an_expression); FORWARD;

    PROCEDURE rec_proc_ref(VAR ok: Boolean; VAR r: a_proc_ref);
        VAR id: identifier;
    BEGIN
      IF symbol_kind = sym_id THEN
      BEGIN
        id:=id_value; get_symbol;
        IF symbol_kind = sym_deref THEN
        BEGIN
          r.kind:=proc_deref; r.the_var:=make_proc_var(id); get_symbol
        END
        ELSE
        BEGIN
          r.kind:=named_ref; r.the_ref:=new_process(id, nil, nil, nil, nil, nil)
        END;
        ok:=true
      END
      ELSE
      BEGIN
        get_symbol;
        IF symbol_kind = sym_id THEN
        BEGIN
          r.kind   :=any_ref;
          r.the_any:=make_proc_var(id_value);
          ok:=true;
          get_symbol
        END
        ELSE syntax_error('Missing process variable <identifier>', ok)
      END
    END;

    PROCEDURE rec_factor(VAR ok: Boolean; VAR e: an_expression);

        PROCEDURE rec_not_expr(VAR ok: Boolean; VAR e: an_expression);
        BEGIN
          get_symbol;
          IF symbol_kind IN first_factor THEN rec_factor(ok, e)
          ELSE syntax_error('Missing or invalid <factor>', ok);
          IF ok THEN e:=new_unary_expr(not_op, e)
        END;

        PROCEDURE rec_io_expr(VAR ok: Boolean; VAR e: an_expression);
            VAR s: a_statement;
        BEGIN
          rec_statement(ok, s);
          IF ok THEN e:=new_io_expr(s)
        END;

        PROCEDURE rec_bracketed_expr(VAR ok: Boolean; VAR e: an_expression);
        BEGIN
          get_symbol;
          IF symbol_kind IN first_expression THEN
          BEGIN
            rec_expression(ok, e);
            IF ok THEN
              IF symbol_kind = sym_close THEN get_symbol
              ELSE syntax_error('Missing ")"', ok)
          END
          ELSE syntax_error('Missing or invalid <expression>', ok)
        END;

    BEGIN
      ok:=true;
      IF symbol_kind = sym_constant THEN
      BEGIN
        e:=new_const_expr(const_value); get_symbol
      END
      ELSE
      IF symbol_kind = sym_Bool_const THEN
      BEGIN
        e:=new_Bool_const_expr(Bool_value); get_symbol
      END
      ELSE
      IF symbol_kind = sym_id THEN
      BEGIN
        e:=new_var_expr(make_variable(id_value, ok));
        IF ok THEN get_symbol
      END
      ELSE
      IF symbol_kind = sym_not THEN rec_not_expr(ok, e)
      ELSE
      IF symbol_kind IN [sym_send, sym_receive] THEN rec_io_expr(ok, e)
      ELSE rec_bracketed_expr(ok, e)
    END;

    PROCEDURE rec_term(VAR ok: Boolean; VAR e: an_expression);
        VAR e1: an_expression;
            op: a_binary_op;
    BEGIN
      rec_factor(ok, e);
      WHILE ok AND (symbol_kind IN [sym_multiply, sym_div, sym_mod, sym_and]) DO
      BEGIN
        op:=binary_operator(symbol_kind);
        get_symbol;
        IF symbol_kind IN first_factor THEN rec_factor(ok, e1)
        ELSE syntax_error('Missing or invalid <factor>', ok);
        IF ok THEN e:=new_binary_expr(op, e, e1)
      END
    END;

    PROCEDURE rec_simple_expr(VAR ok: Boolean; VAR e: an_expression);
        VAR e1            : an_expression;
            the_binary_op : a_binary_op;
            the_unary_op  : a_unary_op;
            have_unary    : Boolean;
    BEGIN
      have_unary:=(symbol_kind IN [sym_plus, sym_minus]);
      IF have_unary THEN
      BEGIN
        the_unary_op:=unary_operator(symbol_kind); get_symbol
      END;
      IF symbol_kind IN first_term THEN rec_term(ok, e)
      ELSE syntax_error('Missing or invalid <term>', ok);
      IF ok AND have_unary THEN e:=new_unary_expr(the_unary_op, e);
      WHILE ok AND (symbol_kind IN [sym_plus, sym_minus, sym_or]) DO
      BEGIN
        the_binary_op:=binary_operator(symbol_kind);
        get_symbol;
        IF symbol_kind IN first_term THEN rec_term(ok, e1)
        ELSE syntax_error('Missing or invalid <term>', ok);
        IF ok THEN e:=new_binary_expr(the_binary_op, e, e1)
      END
    END;

    PROCEDURE rec_expression(* VAR ok: Boolean; VAR e: an_expression *);
        VAR e1: an_expression;
            op: a_binary_op;
    BEGIN
      rec_simple_expr(ok, e);
      WHILE ok AND (symbol_kind IN [sym_eq, sym_neq, sym_ls, sym_leq, sym_gtr, sym_geq]) DO
      BEGIN
        op:=binary_operator(symbol_kind);
        get_symbol;
        IF symbol_kind IN first_simple_expr THEN rec_simple_expr(ok, e1)
        ELSE syntax_error('Missing or invalid <simple_expr>', ok);
        IF ok THEN e:=new_binary_expr(op, e, e1)
      END
    END;

    PROCEDURE rec_stmt_list(VAR ok: Boolean; VAR l: stmt_list); FORWARD;

    PROCEDURE rec_assign(VAR ok: Boolean; VAR s: a_statement);
        VAR variable: a_variable;
            value   : an_expression;
    BEGIN
      variable:=make_variable(id_value, ok);
      IF ok THEN
      BEGIN
        get_symbol;
        ok:=true;
        IF symbol_kind = sym_becomes THEN get_symbol
        ELSE syntax_error('Missing := or invalid variable', ok);
        IF ok THEN
          IF symbol_kind IN first_expression THEN rec_expression(ok, value)
          ELSE syntax_error('Missing or invalid <expression>', ok);
        IF ok THEN s:=new_assignment(variable, value)
      END
    END;

    PROCEDURE rec_send(VAR ok: Boolean; VAR s: a_statement);
        VAR value      : an_expression;
            destination: a_proc_ref;
    BEGIN
      get_symbol;
      ok:=true;
      IF symbol_kind IN first_expression THEN rec_expression(ok, value)
      ELSE value:=new_gen_expr;
      IF ok THEN
        IF symbol_kind = sym_to THEN
        BEGIN
          get_symbol;
          IF symbol_kind IN [sym_any, sym_id] THEN rec_proc_ref(ok, destination)
          ELSE syntax_error('Missing <process reference>', ok)
        END
        ELSE syntax_error('Missing TO or invalid <expression>', ok);
      IF ok THEN s:=new_send(value, destination)
    END;

    PROCEDURE rec_receive(VAR ok: Boolean; VAR s: a_statement);
        VAR variable: a_variable;
            source  : a_proc_ref;
    BEGIN
      get_symbol;
      ok:=true;
      IF symbol_kind = sym_id THEN
      BEGIN
        variable:=make_variable(id_value, ok);
        IF ok THEN get_symbol
      END
      ELSE variable:=new_variable('');
      IF ok THEN
        IF symbol_kind = sym_from THEN
        BEGIN
          get_symbol;
          IF symbol_kind IN [sym_any, sym_id] THEN rec_proc_ref(ok, source)
          ELSE syntax_error('Missing <process reference>', ok)
        END
        ELSE syntax_error('Missing FROM or invalid <variable>', ok);
      IF ok THEN s:=new_receive(variable, source)
    END;

    PROCEDURE rec_if(VAR ok: Boolean; VAR s: a_statement);
        VAR condition: an_expression;
            then_part: a_statement;
            else_part: a_statement;
    BEGIN
      get_symbol;
      IF symbol_kind IN first_expression THEN
      BEGIN
        rec_expression(ok, condition);
        IF ok THEN
          IF symbol_kind = sym_then THEN
          BEGIN
            get_symbol;
            rec_statement(ok, then_part);
            IF ok THEN
              IF symbol_kind = sym_else THEN
              BEGIN
                get_symbol;
                rec_statement(ok, else_part)
              END
              ELSE else_part:=new_empty_stmt
          END
          ELSE syntax_error('Missing THEN or invalid <expression>', ok)
      END
      ELSE syntax_error('Missing or invalid <expression>', ok);
      IF ok THEN s:=new_if(condition, then_part, else_part)
    END;

    PROCEDURE rec_while(VAR ok: Boolean; VAR s: a_statement);
        VAR condition: an_expression;
            statement: a_statement;
    BEGIN
      get_symbol;
      IF symbol_kind IN first_expression THEN
      BEGIN
        rec_expression(ok, condition);
        IF ok THEN
          IF symbol_kind = sym_do THEN
          BEGIN
            get_symbol;
            rec_statement(ok, statement)
          END
          ELSE syntax_error('Missing DO or invalid <expression>', ok)
      END
      ELSE syntax_error('Missing or invalid <expression>', ok);
      IF ok THEN s:=new_while(condition, statement)
    END;

    PROCEDURE rec_compound(VAR ok: Boolean; VAR s: a_statement);
        VAR l: stmt_list;
    BEGIN
      get_symbol;
      rec_stmt_list(ok, l);
      IF ok THEN
        IF symbol_kind = sym_end THEN get_symbol
        ELSE syntax_error('Missing END or ";" or invalid <statement>', ok);
      IF ok THEN s:=new_compound(l)
    END;

    PROCEDURE rec_statement(* VAR ok: Boolean; VAR s: a_statement *);
    BEGIN
      IF symbol_kind IN first_statement THEN
        CASE symbol_kind OF
          sym_begin   : rec_compound(ok, s);
          sym_send    : rec_send    (ok, s);
          sym_receive : rec_receive (ok, s);
          sym_if      : rec_if      (ok, s);
          sym_while   : rec_while   (ok, s);
          sym_id      : rec_assign  (ok, s)
        END
      ELSE
      BEGIN
        ok:=true; s:=new_empty_stmt
      END
    END;

    PROCEDURE rec_stmt_list(* VAR ok: Boolean; VAR l: stmt_list *);
        VAR s: a_statement;
    BEGIN
      l:=nil;
      rec_statement(ok, s);
      IF ok THEN
      BEGIN
        IF s^.kind <> empty_stmt THEN add_statement(s, l);
        WHILE ok AND (symbol_kind = sym_semi) DO
        BEGIN
          get_symbol;
          rec_statement(ok, s);
          IF ok THEN
            IF s^.kind <> empty_stmt THEN add_statement(s, l)
        END
      END;
      IF ok AND (l = nil) THEN add_statement(new_empty_stmt, l)
    END;

    PROCEDURE rec_var_dec(VAR ok: Boolean; VAR variables: var_list);
    BEGIN
      variables:=nil;
      ok:=true;
      IF symbol_kind = sym_var THEN
        BEGIN
        get_symbol;
        IF symbol_kind = sym_id THEN
        BEGIN
          add_variable(new_variable(id_value), variables);
          get_symbol;
          WHILE ok AND (symbol_kind = sym_comma) DO
          BEGIN
            get_symbol;
            IF symbol_kind = sym_id THEN
            BEGIN
              add_variable(new_variable(id_value), variables);
              get_symbol
            END
            ELSE syntax_error('Missing variable <identifier>', ok)
          END;
          IF ok THEN
            IF symbol_kind = sym_semi THEN get_symbol
            ELSE syntax_error('Missing "," or ";"', ok)
        END
        ELSE syntax_error('Missing variable <identifier>', ok)
      END
    END;

    PROCEDURE rec_proc_list(VAR ok: Boolean; VAR l: process_list);
    BEGIN
      l:=nil;
      add_process(new_process(id_value, nil, nil, nil, nil, nil), l);
      get_symbol;
      ok:=true;
      WHILE ok AND (symbol_kind = sym_comma) DO
      BEGIN
        get_symbol;
        IF symbol_kind = sym_id THEN
        BEGIN
          add_process(new_process(id_value, nil, nil, nil, nil, nil), l);
          get_symbol
        END
        ELSE syntax_error('Missing process <identifier>', ok)
      END
    END;

    PROCEDURE rec_connections(VAR ok: Boolean; VAR inputs, outputs: process_list);
    BEGIN
      inputs:=nil; outputs:=nil;
      get_symbol;
      IF symbol_kind = sym_in THEN
      BEGIN
        get_symbol;
        IF symbol_kind = sym_id THEN rec_proc_list(ok, inputs)
        ELSE syntax_error('Missing process <identifier> list', ok)
      END
      ELSE ok:=true;
      IF ok THEN
        IF symbol_kind = sym_out THEN
        BEGIN
          get_symbol;
          IF symbol_kind = sym_id THEN rec_proc_list(ok, outputs)
          ELSE syntax_error('Missing process <identifier> list', ok)
        END;
      IF ok THEN
        IF symbol_kind = sym_close THEN get_symbol
        ELSE syntax_error('Missing ")" in process <identifier> list', ok)
    END;

    PROCEDURE rec_process(VAR ok: Boolean; VAR p: a_process);
        VAR name   : identifier;
            body   : stmt_list;
            inputs : process_list;
            outputs: process_list;
    BEGIN
      inputs:=nil; outputs:=nil; process_vars:=nil;
      get_symbol;
      IF symbol_kind = sym_id THEN
      BEGIN
        name:=id_value;
        get_symbol;
        IF symbol_kind = sym_open THEN rec_connections(ok, inputs, outputs);
        IF ok THEN
          IF symbol_kind = sym_semi THEN
          BEGIN
            get_symbol;
            rec_var_dec(ok, variables);
            IF ok THEN
              IF symbol_kind = sym_begin THEN
              BEGIN
                get_symbol;
                rec_stmt_list(ok, body);
                IF ok THEN
                  IF symbol_kind = sym_end THEN get_symbol
                  ELSE syntax_error('Missing END or ";" or invalid <statement>', ok)
              END
              ELSE syntax_error('Missing BEGIN', ok)
          END
          ELSE syntax_error('Missing ";"', ok)
      END
      ELSE syntax_error('Missing process <identifier>', ok);
      IF ok THEN p:=new_process(name, inputs, outputs, variables, process_vars, body)
    END;

    PROCEDURE rec_program(* name: string; VAR ok: Boolean; VAR prog: a_program *);
        CONST for_UIMS = true;
        VAR   p        : a_process;
    BEGIN
      IF for_UIMS THEN
      BEGIN
        (*
        wi_new_window(w, 0, 2, 65, 7, true);
        wi_flash_string('Syntax Analysis...'); writeln;
        writeln
        *)
      END;
        init_syntax_analysis;
        open_source(name, ok);
        IF ok THEN
        BEGIN
          prog:=nil;
          IF symbol_kind = sym_process THEN rec_process(ok, p)
          ELSE syntax_error('Missing PROCESS', ok);
          IF ok THEN add_process(p, prog);
          WHILE ok AND (symbol_kind = sym_semi) DO
          BEGIN
            get_symbol;
            IF symbol_kind = sym_process THEN rec_process(ok, p)
            ELSE syntax_error('Missing PROCESS', ok);
            IF ok THEN add_process(p, prog)
          END;
          IF ok THEN
            IF symbol_kind = sym_dot THEN get_symbol ELSE syntax_error('Missing "."', ok);
          close_source
        END
        ELSE syntax_error(concat('Unable to find program ', name), ok);
      IF for_UIMS THEN (* wi_dispose_window(true) *)
    END;

END.


